//
//
//	3d2 files parser class
//
//											(C) 1997 JoOl


#include "Parser3d2.h"


Parser3d2::Parser3d2(bool V, int32 BufLen, int32 LexmLen) : ParserBin(V)
{
	Lexr = new LexerBin(BufLen, LexmLen);
}


Parser3d2::~Parser3d2()
{
	delete Lexr;
}


bool
Parser3d2::CheckType()
{
	ASSERT(ParseHeader());
	ASSERT(Header.Sig == SIG_3D2);

	return true;
}


bool
Parser3d2::SetInput(entry_ref& _e)
{
	return inherited::SetInput(_e, P_BIG_ENDIAN);
}


bool
Parser3d2::Parse()
{
	ASSERT(ParseHeader());
	ASSERT(Header.Sig == SIG_3D2);

	PRINT1("%d objects\n", Header.NbObj);
	PRINT5("Light A: %d, bright %d, (%d %d %d)\n", Header.LightAon,
			Header.BrightA, Header.Ax, Header.Ay, Header.Az);
	PRINT5("Light B: %d, bright %d, (%d %d %d)\n", Header.LightBon,
			Header.BrightB, Header.Bx, Header.By, Header.Bz);
	PRINT5("Light C: %d, bright %d, (%d %d %d)\n", Header.LightCon,
			Header.BrightC, Header.Cx, Header.Cy, Header.Cz);

	for (uint16 i = 0; i < Header.NbObj; i++)
	{
		PRINT1("Object %d: ", i);
		if (!ParseObject())
			return false;
	}

	Terminate();

	printf("Done...\n");
	return true;
}


bool
Parser3d2::ParseHeader()
{
	READ_INT16(Header.Sig);
	READ_INT16(Header.NbObj);
	READ_INT16(Header.LightAon);	READ_INT16(Header.LightBon);	READ_INT16(Header.LightCon);
	READ_INT16(Header.BrightA);		READ_INT16(Header.BrightB);		READ_INT16(Header.BrightC);
	READ_INT16(Header.BrightAmb);
	READ_INT16(Header.Az);	READ_INT16(Header.Ay);	READ_INT16(Header.Ax);
	READ_INT16(Header.Bz);	READ_INT16(Header.By);	READ_INT16(Header.Bx);
	READ_INT16(Header.Cz);	READ_INT16(Header.Cy);	READ_INT16(Header.Cx);
	for (uint16 i = 0; i < 32; i++)
		READ_INT16(Header.ObjColor[i]);
	for (uint16 i = 0; i < 32; i++)
		READ_INT16(Header.ColGroup[i]);
	READ_INT16(Header.PalType);
	READ_INT16(Header.WireColor);
	READ_INT16(Header.OutlineColor);
	SKIP_BYTES(86);

	return true;
}



bool
Parser3d2::ParseObject()
{
	char			ObjName[10];
	READ_BYTES(9, ObjName);
	PRINT1("'%s'\n", ObjName);

	ASSERT(ParseVertices());
	ASSERT(ParseFaces());
}



bool
Parser3d2::ParseVertices()
{
	int16			NbVertices;
	READ_INT16(NbVertices);
	PRINT1("%d vertices\n", NbVertices);

	for (int16 i = 0; i < NbVertices; i++)
	{
		int16		_x, _y, _z;
		READ_INT16(_x);
		READ_INT16(_y);
		READ_INT16(_z);

		PRINT4("v%d: %f %f %f\n", i, float(_x)*0.01, float(_y)*0.01, float(_z)*0.01);
	}
	return true;
}


bool
Parser3d2::ParseFaces()
{
	int16			NbFaces;
	READ_INT16(NbFaces);
	PRINT1("%d faces\n", NbFaces);

	for (int16 i = 0; i < NbFaces; i++)
	{
		uint16		Indexes[4];
		READ_INT16(Indexes[0]);
		READ_INT16(Indexes[1]);
		READ_INT16(Indexes[2]);
		READ_INT16(Indexes[3]);

		PRINT6("f%5d: %4d %4d %4d color:%1d, %1x\n", i, Indexes[0], Indexes[1], Indexes[2], Indexes[3] >> 8, Indexes[3] & 0x7);
	}
	return true;
}


