//
//
//	3ds files parser class
//
//											(C) 1997 JoOl


#include "Parser3ds.h"



Parser3ds::Parser3ds(bool V, int32 BufLen, int32 LexmLen) : ParserBin(V)
{
	Lexr = new LexerBin(BufLen, LexmLen);
}


Parser3ds::~Parser3ds()
{
	delete Lexr;
}


bool
Parser3ds::CheckType()
{
	off_t			FileSiz;
	Lexr->GetInputSize(FileSiz);

	READ_CHUNK;
	ASSERT(ChunkId == MAIN_3DS);
	ASSERT(ChunkLen == FileSiz);

	return true;
}


bool
Parser3ds::SetInput(entry_ref& _e)
{
	return inherited::SetInput(_e, P_LITTLE_ENDIAN);
}


bool
Parser3ds::SkipChunk(uint16 ChunkId, uint32 ChunkLen)
{
	SKIP_BYTES(ChunkLen-6);
	INFO("skip chunk");
	return true;
}


bool
Parser3ds::Parse()
{
	READ_CHUNK;

	ASSERT(ChunkId == MAIN_3DS);

	INFO("3ds main");

	CALL(ReadMainChunk);

	Terminate();

	uint32		CurrLen = ChunkLen;
	uint32		Len = ChunkLen;
	CHECK("parsed");
}


bool
Parser3ds::ReadMainChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == EDIT3DS)
		{
			INFO("edit");
			CALL(ReadEditChunk);
		}

		else if (ChunkId == KEYF3DS)
		{
			INFO("keyf");
			CALL(ReadKeyfChunk);
		}

		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("main");
}



bool
Parser3ds::ReadKeyfChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	while (CurrLen < Len)
	{
		READ_CHUNK;

//KEYF_FRAMES
//KEYF_OBJDESCR
		if (ChunkId == EDIT_VIEW1)
		{
			INFO("keyf view1");
			CALL(ReadViewChunk);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("keyf");
}


bool
Parser3ds::ReadEditChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == EDIT_MATERIAL)
		{
			INFO("material");
			CALL(ReadEditMaterialChunk);
		}
		else if (ChunkId == VIEWPORT_TOP)
		{
			uint32		Version;
			READ_INT32(Version);
			PRINT1("## Version %d\n", Version);
		}
		else if (ChunkId == EDIT_VIEW1)
		{
			INFO("edit view1");
			CALL(ReadViewChunk);
		}
		else if (ChunkId == EDIT_BACKGR)
		{
			INFO("background");
			CALL(ReadBackgroundChunk);
		}
		else if (ChunkId == EDIT_AMBIENT)
		{
			INFO("ambient");
			CALL(ReadAmbientChunk);
		}
		else if (ChunkId == EDIT_OBJECT)
		{
			INFO("object");
			CALL(ReadObjectChunk);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("edit");
}


bool
Parser3ds::ReadEditMaterialChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == MAT_NAME)
		{
			char		Name[16];
			READ_STRING(16, Name);
			PRINT1("material name '%s'\n", Name);
		}

//MAT_AMBIENTCOLOR
//MAT_DIFFUSECOLOR
//MAT_SPECULARCOLOR
//MAT_SHININESS
//MAT_TRANSPARENCY
//MAT_TRANSFALLOFF
//MAT_REFLECTBLUR
//MAT_TEXTUREMAP
//MAT_OPACITYMAP
//MAT_TRANSFOFFENABLE
//MAT_REFLBLURENABLE
//MAT_BUMPMAP
//MAT_MAPFILE

		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("edit material");
}


bool
Parser3ds::ReadViewChunk(uint32 Len)
{
	uint32			CurrLen = 6;

	// 14 found from experimentation and observation (HexJuggler)
	SkipBytes(14);
	CurrLen += 14;

	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == EDIT_VIEW_P1)
		{
			INFO("view p1");
			CALL(ReadViewportChunk);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("view");
}


bool
Parser3ds::ReadViewportChunk(uint32 Len)
{
	uint16			CurrLen = 6;

	uint16			Attrib;
	READ_INT16(Attrib);
	SKIP_BYTES(5*sizeof(uint16));
	uint16			Port;
	READ_INT16(Port);
	CurrLen += (1+5+1) * sizeof(uint16);

	if (Port == VIEWPORT_CAMERA)
	{
		SKIP_BYTES(12*sizeof(uint16));
		CurrLen += 12 * sizeof(uint16);

		char		Name[16];
		READ_STRING(16, Name);
		CurrLen += strlen(Name)+1;
		Port = 8-1;

		// Gruik !!!
		SKIP_BYTES(2);
		CurrLen += 2;

		PRINT1("camera '%s'\n", Name);
	}

	// Gruik !!!
	if (CurrLen < Len)
	{
		SKIP_BYTES(Len - CurrLen);
		CurrLen += Len-CurrLen;
	}

	CHECK("viewport");
}


bool
Parser3ds::ReadAmbientChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == COLOR_RGBFLOAT)
		{
			INFO("rgb float");
			CALL(ReadRgbFloatChunk);
		}
		else if (ChunkId == COLOR_RGBBYTE)
		{
			INFO("rgb byte");
			CALL(ReadRgbByteChunk);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("ambient");
}


bool
Parser3ds::ReadRgbFloatChunk(uint32 Len)
{
	uint32		CurrLen = 6;
	float		r, g, b;
	READ_FLOAT(r);
	READ_FLOAT(g);
	READ_FLOAT(b);
	CurrLen += 3*sizeof(float);
	CHECK("rgbf");
}


bool
Parser3ds::ReadRgbByteChunk(uint32 Len)
{
	uint32		CurrLen = 6;
	uint8		r, g, b;
	READ_INT8(r);
	READ_INT8(g);
	READ_INT8(b);
	CurrLen += 3*sizeof(uint8);
	CHECK("rgbb");
}


bool
Parser3ds::ReadBackgroundChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == COLOR_RGBFLOAT)
		{
			INFO("rgb float");
			CALL(ReadRgbFloatChunk);
		}
		else if (ChunkId == COLOR_RGBBYTE)
		{
			INFO("rgb byte");
			CALL(ReadRgbByteChunk);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("background");
}


bool
Parser3ds::ReadObjectChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	char			Name[16];
	READ_STRING(16, Name);
	CurrLen += strlen(Name)+1;

	PRINT1("object '%s'\n", Name);

	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == OBJ_TRIMESH)
		{
			INFO("tri mesh");
			CALL(ReadTriMeshChunk);
		}
		else if (ChunkId == OBJ_LIGHT)
		{
			INFO("light");
			CALL(ReadLightChunk);
		}
		else if (ChunkId == OBJ_CAMERA)
		{
			INFO("camera");
			CALL(ReadCameraChunk);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("object");
}


bool
Parser3ds::ReadTriMeshChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == TRI_VERTLIST)
		{
			INFO("vertices");
			CALL(ReadVerticesChunk);
		}
		else if (ChunkId == TRI_FACELIST)
		{
			INFO("faces");
			CALL(ReadFacesChunk);
		}
		else if (ChunkId == TRI_LOCALAXIS)
		{
			INFO("transfo");
			CALL(ReadTransfoChunk);
		}
		else if (ChunkId == TRI_VISIBLE)
		{
			INFO("visible");
			uint8		Bool;
			READ_INT8(Bool);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("trimesh");
}


bool
Parser3ds::ReadTransfoChunk(uint32 Len)
{
	uint32		CurrLen = 6;
	float		u[3];
	float		v[3];
	float		n[3];
	float		t[3];

	READ_FLOAT(u[0]);
	READ_FLOAT(u[1]);
	READ_FLOAT(u[2]);

	READ_FLOAT(v[0]);
	READ_FLOAT(v[1]);
	READ_FLOAT(v[2]);

	READ_FLOAT(n[0]);
	READ_FLOAT(n[1]);
	READ_FLOAT(n[2]);

	READ_FLOAT(t[0]);
	READ_FLOAT(t[1]);
	READ_FLOAT(t[2]);

	CurrLen += 12 * sizeof(float);

	PRINT4("%f %f %f %f\n", u[0], v[0], n[0], 0.0);
	PRINT4("%f %f %f %f\n", u[1], v[1], n[1], 0.0);
	PRINT4("%f %f %f %f\n", u[2], v[2], n[2], 0.0);
	PRINT4("%f %f %f %f\n", t[0], t[1], t[2], 1.0);

	CHECK("transfo");
}


bool
Parser3ds::ReadFacesChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	uint16			NbFaces;
	READ_INT16(NbFaces);
	CurrLen += 2;
	for (uint16 i = 0; i < NbFaces; i++)
	{
		uint16		i0, i1, i2, flags;
		READ_INT16(i0);
		READ_INT16(i1);
		READ_INT16(i2);
		READ_INT16(flags);
		PRINT6("%4d:%4d %4d %4d %c %4x\n", i, i0, i1, i2, (flags & 0x8)?'m':'f', flags & 0x7);
	}
	CurrLen += 4*sizeof(uint16)*NbFaces;

	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == TRI_MATERIAL)
		{
			INFO("material");
			CALL(ReadFaceMaterialChunk);
		}
		else if (ChunkId == TRI_SMOOTH)
		{
			INFO("faces smooth");
			CALL2(ReadFaceSmoothChunk, NbFaces);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}

	CHECK("faces");
}


bool
Parser3ds::ReadFaceMaterialChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	char			Name[16];
	READ_STRING(16, Name);
	CurrLen += strlen(Name) + 1;

	// Number of faces concerned with the current material
	uint16			NbFaces;
	READ_INT16(NbFaces);
	CurrLen += 2;
	uint16			Tmp = 0;
	for (uint16 i = 0; i < NbFaces; i++)
	{
		if ((Verbose) && (Tmp == 0))
			printf("%s", Indent);
		uint16		FaceIndex;
		READ_INT16(FaceIndex);
		if (Verbose)
			printf("%4d ", FaceIndex);
		Tmp++;
		if ((Verbose) && (Tmp == 10))
		{
			Tmp = 0;
			printf("\n");
		}
	}
	CurrLen += sizeof(uint16) * NbFaces;
	if (Tmp != 0)
		printf("\n");

	// Should/Must be useless
	while (CurrLen < Len)
	{
		READ_CHUNK;
		SKIP;
		CurrLen += ChunkLen;
	}
	CHECK("face material");
}


bool
Parser3ds::ReadVerticesChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	uint16			NbVertices;
	READ_INT16(NbVertices);
	CurrLen += 2;
	for (uint16 i = 0; i < NbVertices; i++)
	{
		float		x, y, z;
		READ_FLOAT(x);
		READ_FLOAT(y);
		READ_FLOAT(z);
		PRINT3("%f %f %f\n", x, y, z);
	}
	CurrLen += 3*sizeof(float) * NbVertices;

	// Should/Must be useless
	while (CurrLen < Len)
	{
		READ_CHUNK;
		SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("vertices");
}


bool
Parser3ds::ReadFaceSmoothChunk(uint32 Len, uint16 NbFaces)
{
	uint16			CurrLen = 6;
	uint16			Tmp = 0;

	for (uint16 i = 0; i < NbFaces; i++)
	{
		if ((Verbose) && (Tmp == 0))
			printf("%s", Indent);

		// the nth bit is set if the face belongs to the nth smoothing group
		uint32		Set;
		READ_INT32(Set);
		if (Verbose)
			printf("%08x ", Set);
		Tmp++;
		if ((Verbose) && (Tmp == 5))
		{
			Tmp = 0;
			printf("\n");
		}
	}
	CurrLen += sizeof(uint32) * NbFaces;
	if ((Verbose) && (Tmp != 0))
		printf("\n");

	// Should/Must be useless
	while (CurrLen < Len)
	{
		READ_CHUNK;
		SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("face smooth");
}


bool
Parser3ds::ReadCameraChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	float			Scalar;
	float			x, y, z;
	READ_FLOAT(x);	READ_FLOAT(y);	READ_FLOAT(z);	// Eye
	READ_FLOAT(x);	READ_FLOAT(y);	READ_FLOAT(z);	// Camera focus
	READ_FLOAT(Scalar);		// Rotation
	READ_FLOAT(Scalar);		// Lens (mm)
	CurrLen += 8*sizeof(float);

	while (CurrLen < Len)
	{
		READ_CHUNK;
		SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("camera");
}


bool
Parser3ds::ReadLightChunk(uint32 Len)
{
	uint32			CurrLen = 6;
	float			x, y, z;

	READ_FLOAT(x);	READ_FLOAT(y);	READ_FLOAT(z);	// Light position
	CurrLen += 3*sizeof(float);
	while (CurrLen < Len)
	{
		READ_CHUNK;

		if (ChunkId == LIGHT_OFF)
		{
			INFO("light on/off");
		}
		else if (ChunkId == LIGHT_SPOT)
		{
			INFO("light spot");
			CALL(ReadSpotChunk);
		}
		else if (ChunkId == COLOR_RGBFLOAT)
		{
			INFO("rgb float");
			CALL(ReadRgbFloatChunk);
		}
		else if (ChunkId == COLOR_RGBBYTE)
		{
			INFO("rgb byte");
			CALL(ReadRgbByteChunk);
		}
		else
			SKIP;

		CurrLen += ChunkLen;
	}
	CHECK("light");
}


bool
Parser3ds::ReadSpotChunk(uint32 Len)
{
	uint32		CurrLen = 6;

	float			x, y, z;
	float			HotSpot, FallOff;

	READ_FLOAT(x);	READ_FLOAT(y);	READ_FLOAT(z);	// Spot position
	READ_FLOAT(HotSpot);
	READ_FLOAT(FallOff);
	CurrLen += 5*sizeof(float);

	while (CurrLen < Len)
	{
		READ_CHUNK;
		SKIP;

		CurrLen += ChunkLen;
	}

	CHECK("spot");
}

