//
//
//	DXF file parser class
//
//											(C) 1997 JoOl


#include "ParserDxf.h"



ParserDxf::ParserDxf(bool V, int32 BufLen, int32 LexmLen): ParserText(V)
{
	Lexr = new LexerDxf(BufLen, LexmLen);
}


ParserDxf::~ParserDxf()
{
	delete Lexr;
}


bool
ParserDxf::CheckType()
{
	MATCH_GROUP(0, LX_TOKEN_SECTION);

	return true;
}


bool
ParserDxf::MatchGroup(int32 Gcode, int32& Out)
{
	MATCH_GROUP(Gcode, LX_TYPE_INTEGER);
	Out = Lexr->GetTokenValueInt();
	return true;
}


bool
ParserDxf::MatchGroup(int32 Gcode, float& Out)
{
	MATCH_GROUP(Gcode, LX_TYPE_FLOAT);
	Out = Lexr->GetTokenValueFloat();
	return true;
}


bool
ParserDxf::MatchGroupCast(int32 Gcode, float& Out)
{
	if (Peek(Gcode, LX_TYPE_INTEGER))
	{
		MATCH_GROUP(Gcode, LX_TYPE_INTEGER);
		Out = float(Lexr->GetTokenValueInt());
		return true;
	}
	MATCH_GROUP(Gcode, LX_TYPE_FLOAT);
	Out = Lexr->GetTokenValueFloat();
	return true;
}


bool
ParserDxf::MatchGroup(int32 Gcode, char*& Out)
{
	MATCH_GROUP(Gcode, LX_TYPE_STRING);
	Out = Lexr->GetTokenValueStr();
	return true;
}


bool
ParserDxf::MatchGroupCast(int32 Gcode, char*& Out)
{
	if (!Peek(Gcode))
		return false;

	if (Peek(Gcode, LX_TYPE_STRING))
	{
		MATCH_GROUP(Gcode, LX_TYPE_STRING);
		Out = Lexr->GetTokenValueStr();
		return true;
	}
	SKIP_GROUP;
	Out = strdup("0");
	return true;
}


void
ParserDxf::PrintGroup()
{
	if (ParserToken == LX_TYPE_INTEGER)
	{
		PRINT2("group %d: %d\n", GroupCode, Lexr->GetTokenValueInt());
	}
	else if (ParserToken == LX_TYPE_FLOAT)
	{
		PRINT2("group %d: %f\n", GroupCode, Lexr->GetTokenValueFloat());
	}
	else if (ParserToken == LX_TYPE_STRING)
	{
		PRINT2("group %d: '%s'\n", GroupCode, Lexr->GetTokenValueStr());
	}
	else
	{
		PRINT2("group %d: %s\n", GroupCode, Lexr->GetTokenLexeme(ParserToken));
	}
}


bool
ParserDxf::NextToken()
{
	if (PeekFlag)
	{
		PeekFlag = false;
		return true;
	}
	bool		Out = Lexr->NextToken(ParserToken);
	GroupCode = ((LexerDxf* )Lexr)->GetGroupCode();
	return Out;
}


bool
ParserDxf::Match(int32 Gcode)
{
	ASSERT(NextToken());
	return (Gcode == GroupCode);
}


bool
ParserDxf::Match(int32 Gcode, uint32 Tokn)
{
	ASSERT(NextToken());
	if ((Gcode == GroupCode) && (Tokn == ParserToken))
		return true;

	if (ParserToken == LX_TYPE_INTEGER)
	{
		PRINT5("### expected [%d %s], got [%d %d (integer %d)]\n",
				Gcode, Lexr->GetTokenLexeme(Tokn), GroupCode,
				ParserToken, Lexr->GetTokenValueInt());
	}
	else if (ParserToken == LX_TYPE_FLOAT)
	{
		PRINT5("### expected [%d %s], got [%d %d (float %f)]\n",
				Gcode, Lexr->GetTokenLexeme(Tokn), GroupCode,
				ParserToken, Lexr->GetTokenValueFloat());
	}
	else if (ParserToken == LX_TYPE_STRING)
	{
		PRINT5("### expected [%d %s], got [%d %d (string '%s')]\n",
				Gcode, Lexr->GetTokenLexeme(Tokn), GroupCode,
				ParserToken, Lexr->GetTokenValueStr());
	}
	else
	{
		PRINT5("### expected [%d %s], got [%d %d (lexeme %s)]\n",
				Gcode, Lexr->GetTokenLexeme(Tokn), GroupCode,
				ParserToken, Lexr->GetTokenLexeme(ParserToken));
	}
	return false;
}


bool
ParserDxf::Peek(int32 Gcode)
{
	if (!PeekFlag)
	{
		PeekFlag = true;
		bool			Out = Lexr->NextToken(ParserToken);
		GroupCode = ((LexerDxf* )Lexr)->GetGroupCode();
		if (!Out)
			return false;
	}
	return (Gcode == GroupCode);
}


bool
ParserDxf::Peek(int32 Gcode, uint32 Tokn)
{
	if (!PeekFlag)
	{
		PeekFlag = true;
		bool			Out = Lexr->NextToken(ParserToken);
		GroupCode = ((LexerDxf* )Lexr)->GetGroupCode();
		if (!Out)
			return false;
	}
	return (Gcode == GroupCode) && (Tokn == ParserToken);
}


bool
ParserDxf::ParseTo(int32 Gcode)
{
	for (;;)
	{
		if (Peek(0, LX_TOKEN_EOF))
			return false;

		if (Peek(Gcode))
			break;

		SKIP_GROUP;
	}
	return true;
}


bool
ParserDxf::ParseTo(int32 Gcode, uint32 Tokn)
{
	for (;;)
	{
		if (Peek(0, LX_TOKEN_EOF))
			return false;

		if (Peek(Gcode, Tokn))
			break;

		SKIP_GROUP;
	}
	return true;
}


bool
ParserDxf::ParseVector(int32 Gcode, float& x, float& y, float& z)
{
	ASSERT(MatchGroupCast(Gcode, x));
	ASSERT(MatchGroupCast(Gcode+10, y));
	ASSERT(MatchGroupCast(Gcode+20, z));
	return true;
}



//-------------------------------------------------
// Main parsing routine
//-------------------------------------------------

bool
ParserDxf::Parse()
{
	for (;;)
	{
		if (Peek(0, LX_TOKEN_EOF))
			break;

		// Parse to next section
		MATCH_GROUP(0, LX_TOKEN_SECTION);
		PrintGroup();

		if (Peek(2, LX_TOKEN_HEADER))
		{
			ASSERT2(ParseHeader(), "header");
		}
		else if (Peek(2, LX_TOKEN_TABLES))
		{
			ASSERT2(ParseTables(), "tables");
		}
		else if (Peek(2, LX_TOKEN_ENTITIES))
		{
			ASSERT2(ParseEntities(), "entities");
		}
		else
		{
			ASSERT(ParseTo(0, LX_TOKEN_ENDSEC));
			MATCH_GROUP(0, LX_TOKEN_ENDSEC);
		}
	}

	Terminate();

	printf("Done...\n");
	return true;
}



//-------------------------------------------------
// Header related routines
//-------------------------------------------------

bool
ParserDxf::ParseHeader()
{
	MATCH_GROUP(2, LX_TOKEN_HEADER);

	while (!Peek(0, LX_TOKEN_EOF) && !Peek(0, LX_TOKEN_ENDSEC))
	{
		if (Peek(9, LX_TOKEN_S_ACADVER))
		{
			ASSERT2(ParseVersion(), "version");
		}
		else
		{
			SKIP_GROUP;
		}
	}
	MATCH_GROUP(0, LX_TOKEN_ENDSEC);
	return true;
}


bool
ParserDxf::ParseVersion()
{
	MATCH_GROUP(9, LX_TOKEN_S_ACADVER);

	MATCH_GROUP(1, LX_TYPE_STRING);

	char*			VersionStr = Lexr->GetTokenValueStr();
	if ((!strcmp(VersionStr, VERSION_R11))
	|| (!strcmp(VersionStr, VERSION_R12)))
	{
		PRINT1("version %s: r11/r12\n", VersionStr);
	}
	else if (!strcmp(VersionStr, VERSION_R10))
	{
		PRINT1("version %s: r10\n", VersionStr);
	}
	else
		PRINT1("version %s: unknown\n", VersionStr);

	return true;
}



//-------------------------------------------------
// Tables related routines
//-------------------------------------------------


bool
ParserDxf::ParseTables()
{
	MATCH_GROUP(2, LX_TOKEN_TABLES);

	while (!Peek(0, LX_TOKEN_EOF) && !Peek(0, LX_TOKEN_ENDSEC))
	{
		if (Peek(0, LX_TOKEN_TABLE))
		{
			ASSERT2(ParseTable(), "table");
		}
		else
		{
			SKIP_GROUP;
		}
	}
	MATCH_GROUP(0, LX_TOKEN_ENDSEC);
	return true;
}


bool
ParserDxf::ParseTable()
{
	int32			TableType;
	int32			Flags;

	MATCH_GROUP(0, LX_TOKEN_TABLE);

	ASSERT(Match(2));
	TableType = ParserToken;

	ASSERT(MatchGroup(70, Flags));

	if (TableType == LX_TOKEN_APPID)
	{
		ASSERT2(ParseAppId(), "appid");
	}
	else if (TableType == LX_TOKEN_LAYER)
	{
		ASSERT2(ParseLayers(), "layers");
	}
	else if (TableType == LX_TOKEN_VIEW)
	{
		ASSERT2(ParseView(), "view");
	}
	else
	{
		ASSERT(ParseTo(0, LX_TOKEN_ENDTAB));
		MATCH_GROUP(0, LX_TOKEN_ENDTAB);
	}
	return true;
}


bool
ParserDxf::ParseAppId()
{
	char*			AppName;
	int32			Flags;


	// Tables can be empty!?!
	if (Peek(0, LX_TOKEN_ENDTAB))
	{
		MATCH_GROUP(0, LX_TOKEN_ENDTAB);
		return true;
	}

	MATCH_GROUP(0, LX_TOKEN_APPID);
	ASSERT(MatchGroup(2, AppName));
	AppName = DetachLexeme(AppName);
	ASSERT(MatchGroup(70, Flags));

	PRINT2("app '%s', flags=%d\n", AppName, Flags);

	return true;
}


bool
ParserDxf::ParseLayers()
{
	while (!Peek(0, LX_TOKEN_EOF) && !Peek(0, LX_TOKEN_ENDTAB))
	{
		if (Peek(0, LX_TOKEN_LAYER))
		{
			ASSERT2(ParseLayer(), "layer");
		}
		else
		{
			SKIP_GROUP;
		}
	}
	MATCH_GROUP(0, LX_TOKEN_ENDTAB);
	return true;
}


bool
ParserDxf::ParseLayer()
{
	int32			LayerNameInt;
	char*			LayerName = NULL;
	int32			Flags;
	int32			Color;
	char*			LayerTypeName;

	MATCH_GROUP(0, LX_TOKEN_LAYER);

	// Names can be integers!!!
	ASSERT(Match(2));
	if (ParserToken == LX_TYPE_INTEGER)
		LayerNameInt = Lexr->GetTokenValueInt();
	else
		LayerName = DetachLexeme(Lexr->GetTokenValueStr());

	ASSERT(MatchGroup(70, Flags));
	ASSERT(MatchGroup(62, Color));
	ASSERT(MatchGroup(6, LayerTypeName));
	LayerTypeName = DetachLexeme(LayerTypeName);

	if (!LayerName)
	{
		PRINT4("layer '%d' flags=%d, color=%d, type='%s'\n", LayerNameInt, Flags,
				Color, LayerTypeName);
	}
	else
	{
		PRINT4("layer '%s' flags=%d, color=%d, type='%s'\n", LayerName, Flags,
				Color, LayerTypeName);
	}

	delete[] LayerName;
	delete[] LayerTypeName;
	return true;
}


bool
ParserDxf::ParseView()
{
	char*			ViewName;
	int32			Flags;
	float			ScreenWidth, ScreenHeight;
	float			CenterX, CenterY;
	float			Vdx, Vdy, Vdz;		// View direction
	float			Tox, Toy, Toz;		// Target point
	float			LensLength;
	float			F, B;				// Front & Back clipping plane distance
	float			TwistAngle;
	int32			ViewMode;

	// Tables can be empty!?!
	if (Peek(0, LX_TOKEN_ENDTAB))
	{
		MATCH_GROUP(0, LX_TOKEN_ENDTAB);
		return true;
	}

	MATCH_GROUP(0, LX_TOKEN_VIEW);

	while (!Peek(0))
	{
		if (Peek(2))
		{
			ASSERT(MatchGroup(2, ViewName));
			ViewName = DetachLexeme(ViewName);
		}
		else if (Peek(70))
		{
			ASSERT(MatchGroup(70, Flags));
		}
		else if (Peek(40))
		{
			ASSERT(MatchGroup(40, ScreenWidth));
		}
		else if (Peek(41))
		{
			ASSERT(MatchGroup(41, ScreenHeight));
		}
		else if (Peek(10))
		{
			ASSERT(MatchGroup(10, CenterX));
		}
		else if (Peek(20))
		{
			ASSERT(MatchGroup(20, CenterY));
		}
		else if (Peek(11))
		{
			ASSERT(ParseVector(11, Vdx, Vdy, Vdz));
		}
		else if (Peek(12))
		{
			ASSERT(ParseVector(12, Tox, Toy, Toz));
		}
		else if (Peek(42))
		{
			ASSERT(MatchGroup(42, LensLength));
		}
		else if (Peek(43))
		{
			ASSERT(MatchGroup(43, F));
		}
		else if (Peek(44))
		{
			ASSERT(MatchGroup(44, B));
		}
		else if (Peek(50))
		{
			ASSERT(MatchGroup(50, TwistAngle));
		}
		else if (Peek(71))
		{
			ASSERT(MatchGroup(71, ViewMode));	// No information on bits in "view mode"
		}
	}

	PRINT6("view '%s': flags=%d, screen=%fx%f center=%fx%f\n", ViewName,
			Flags, ScreenWidth, ScreenHeight, CenterX, CenterY);
	PRINT6("  dir (%f %f %f) to (%f %f %f)\n", Vdx, Vdy, Vdz, Tox, Toy, Toz);
	PRINT5("  lens=%f F=%f B=%f twist=%f view mode=%d\n", LensLength,
			F, B, TwistAngle, ViewMode);

	return true;
}


//-------------------------------------------------
// Entities related routines
//-------------------------------------------------


bool
ParserDxf::ParseEntities()
{
	MATCH_GROUP(2, LX_TOKEN_ENTITIES);

	while (!Peek(0, LX_TOKEN_EOF))
	{
		if (Peek(0, LX_TOKEN_LINE))
		{
			ASSERT2(ParseLine(), "line");
		}
		else if (Peek(0, LX_TOKEN_POINT))
		{
			ASSERT2(ParsePoint(), "point");
		}
		else if (Peek(0, LX_TOKEN_VERTEX))
		{
			ASSERT2(ParseVertex(), "vertex");
		}
		else if (Peek(0, LX_TOKEN_3DFACE))
		{
			ASSERT2(ParseFace3d(), "face3d");
		}
		else if (Peek(0, LX_TOKEN_POLYLINE))
		{
			ASSERT2(ParsePolyline(), "polyline");
		}
		else
		{
			SKIP_GROUP;
		}
	}

	return true;
}


bool
ParserDxf::ParseLine()
{
	char*			LayerName;
	float			xs, ys, zs;
	float			xe, ye, ze;

	MATCH_GROUP(0, LX_TOKEN_LINE);

	while (!Peek(0))
	{
		if (Peek(10))
		{
			ASSERT(ParseVector(10, xs, ys, zs));
		}
		else if (Peek(11))
		{
			ASSERT(ParseVector(11, xe, ye, ze));
		}
		else if (Peek(8))
		{
			ASSERT(MatchGroupCast(8, LayerName));
			LayerName = DetachLexeme(LayerName);
		}
		else
		{
			SKIP_GROUP;
		}
	}
	PRINT7("line in '%s': (%f %f %f) to (%f %f %f)\n", LayerName,
			xs, ys, zs, xe, ye, ze);
	delete[] LayerName;

	return true;
}


bool
ParserDxf::ParsePoint()
{
	char*			LayerName;
	float			x, y, z;

	MATCH_GROUP(0, LX_TOKEN_POINT);

	while (!Peek(0))
	{
		if (Peek(10))
		{
			ASSERT(ParseVector(10, x, y, z));
		}
		else if (Peek(8))
		{
			ASSERT(MatchGroupCast(8, LayerName));
			LayerName = DetachLexeme(LayerName);
		}
		else
		{
			SKIP_GROUP;
		}
	}
	PRINT4("point in '%s': (%f %f %f)\n", LayerName, x, y, z);
	delete[] LayerName;

	return true;
}


bool
ParserDxf::ParseVertex()
{
	char*			LayerName;
	float			x, y, z;
	int32			Flags;

	MATCH_GROUP(0, LX_TOKEN_VERTEX);

	while (!Peek(0))
	{
		if (Peek(10))
		{
			ASSERT(ParseVector(10, x, y, z));
		}
		else if (Peek(8))
		{
			ASSERT(MatchGroupCast(8, LayerName));
			LayerName = DetachLexeme(LayerName);
		}
		else if (Peek(70))
		{
			ASSERT(MatchGroup(70, Flags));
		}
		else
		{
			SKIP_GROUP;
		}
	}
	PRINT4("vertex in '%s': (%f %f %f)\n", LayerName, x, y, z);
	delete[] LayerName;

	return true;
}


bool
ParserDxf::ParseFace3d()
{
	char*			LayerName = NULL;
	float			x0, y0, z0;
	float			x1, y1, z1;
	float			x2, y2, z2;
	float			x3, y3, z3;
	int32			Flags = 0;

	MATCH_GROUP(0, LX_TOKEN_3DFACE);

	while (!Peek(0))
	{
		if (Peek(10))
		{
			ASSERT(ParseVector(10, x0, y0, z0));
		}
		else if (Peek(11))
		{
			ASSERT(ParseVector(11, x1, y1, z1));
		}
		else if (Peek(12))
		{
			ASSERT(ParseVector(12, x2, y2, z2));
		}
		else if (Peek(13))
		{
			ASSERT(ParseVector(13, x3, y3, z3));
		}
		else if (Peek(8))
		{
			ASSERT(MatchGroupCast(8, LayerName));
			LayerName = DetachLexeme(LayerName);
		}
		else if (Peek(70))
		{
			ASSERT(MatchGroup(70, Flags));
		}
		else
		{
			SKIP_GROUP;
		}
	}
	PRINT5("face3d in '%s', flags=%d: (%f %f %f)\n", LayerName, Flags, x0, y0, z0);
	PRINT6("  (%f %f %f) (%f %f %f)\n", x1, y1, z1, x2, y2, z2);
	if ((x2 != x3) || (y2 != y3) || (z2 != z3))
	{
		PRINT3("  (%f %f %f) %x\n", x3, y3, z3);
	}
	delete[] LayerName;

	return true;
}


bool
ParserDxf::ParsePolyline()
{
	char*			LayerName;
	int32			VerticesFollow;
	float			x = 0, y = 0, z = 0;
	int32			Flags;

	MATCH_GROUP(0, LX_TOKEN_POLYLINE);

	// Parse header first
	while (!Peek(0))
	{
		if (Peek(10))
		{
			ASSERT(ParseVector(10, x, y, z));
		}
		else if (Peek(66))
		{
			ASSERT(MatchGroup(66, VerticesFollow));
		}
		else if (Peek(8))
		{
			ASSERT(MatchGroupCast(8, LayerName));
			LayerName = DetachLexeme(LayerName);
		}
		else if (Peek(70))
		{
			ASSERT(MatchGroup(70, Flags));
		}
		else
		{
			SKIP_GROUP;
		}
	}
	PRINT6("polyline in '%s', flags=%d: vert. follow:%d (%f %f %f)\n",
			LayerName, Flags, VerticesFollow, x, y, z);
	delete[] LayerName;

	while (!Peek(0, LX_TOKEN_SEQEND))
	{
		if (Peek(0, LX_TOKEN_VERTEX))
		{
			ASSERT(ParseVertex());
		}
		else
		{
			SKIP_GROUP;
		}
	}
	MATCH_GROUP(0, LX_TOKEN_SEQEND);

	return true;
}

