//
//
//	Basic text-file lexer class
//
//											(C) 1997 JoOl



#include "LexerBasic.h"



LexerBasic::LexerBasic(int32 BufLen, int32 LexmLen): LexerText(BufLen, LexmLen)
{
	SetCharset();
	SetLexemes();
	SetSymbols();		// Symbols are not known a-priori, this is just an example
}


LexerBasic::~LexerBasic()
{
}


void
LexerBasic::SetCharset()
{
	SetDefaultCharset();

	SetChar('-', LX_CHAR_MINUS);
	SetChar(0x0d, LX_CHAR_SPACE);
}


void
LexerBasic::SetLexemes()
{
	// Well, nothing much...
	SortLexemes();
}


void
LexerBasic::SetSymbols()
{
	// We don't actually care about neither the type, nor the value
	AddSymbol(new Symbol("root", LX_KIND_SCALAR_VAR, int32(0)));
	AddSymbol(new Symbol("ltibia", LX_KIND_SCALAR_VAR, int32(0)));
	AddSymbol(new Symbol("rfemur", LX_KIND_SCALAR_VAR, int32(0)));
	AddSymbol(new Symbol("rfoot", LX_KIND_SCALAR_VAR, int32(0)));
	AddSymbol(new Symbol("lowerback", LX_KIND_SCALAR_VAR, int32(0)));

	// Well, there are some other examples, but it's useless, Huh?
	SortSymbols();
}


bool
LexerBasic::SetInput(entry_ref& Ref)
{
	return inherited::SetInput(Ref);
}



bool
LexerBasic::NextToken(uint32& Out)
{
	InitNextToken();

	// This loop is exited when a lexeme is found
	for (;;)
	{
		if (EndOfFile)
		{
			Out = Token = LX_CHAR_EOF;
			return true;
		}

		LexmStart = LexmEnd;
		uint8			CharType = Charset[uint8(*LexmEnd)];

		if ((CharType == LX_CHAR_DIGIT) || (*LexmEnd == '.'))
		{
			if (!LexNumber())
				return false;
			Out = Token;
			return true;
		}
		else if (CharType == LX_CHAR_ALPHA)
		{
			if (!LexString())
				return false;
			Out = Token;
			return true;
		}
		else if (CharType == LX_CHAR_MINUS)
		{
			FWD;
			if (!LexNumber())
				return false;
			Out = Token;
			return true;
		}
		else if (CharType == LX_CHAR_SPACE)
		{
			if (!SkipSpaces())
				return false;
		}
		else if (CharType == LX_CHAR_ILLEGAL)	// Illegal characters generate an error
		{
			Out = Token = LX_CHAR_ILLEGAL;
			IllegalCharacter = *LexmEnd;
			return false;
		}
		else		// lexeme?
		{
			Out = Token = CharType;
			return true;
		}
	}
}
